/**
 * Adapted from:
 * Copyright (c) 2022 Raspberry Pi (Trading) Ltd.
 * SPDX-License-Identifier: BSD-3-Clause
 * the ntp routines are just copied into this code
 * 
 * Integrate LWIP with protothreads
 * Threads:
 * --get network time
 * --blink cyw43 LED slowly
 * --toggle gpio2 fast (few KHz)
 * 
 * From the hardware design doc: "The chosen crystal is an ABLS-12.000MHZ-B4-T (Y1 in Figure 8). 
 * This is a commonly available 12MHzvcrystal with a 30ppm tolerance."
 * 30 ppm equates to 3e-5. since there are 86400 sec/day, the means about 2-3 sec/day of rtc drift.
 * Therefore need to read network time at least once/day
 */

// =====================================
// network includes
#include <string.h>
#include <time.h>
#include <pico/multicore.h>
#include "hardware/sync.h"
#include "hardware/uart.h"
#include "hardware/rtc.h"
#include "pico/util/datetime.h"

#include "pico/stdlib.h"
#include "pico/cyw43_arch.h"

#include "lwip/dns.h"
#include "lwip/pbuf.h"
#include "lwip/udp.h"

//========================================
// necessary to connect to wireless
// !!! Do NOT post this info !!!
#define WIFI_SSID "*****"
#define WIFI_PASSWORD "******"

//========================================
// protothreads
#include "pt_cornell_rp2040_v1_1_2.h"
// signalled by serial thread, triggers ntp thread
struct pt_sem ntp_go_s ;
int set_rtc_flag = 0 ;

// time strutures for RTC
char datetime_buf[256];
char *datetime_str = &datetime_buf[0];

#define EDT  -4
#define EST  -5
int zone = EDT ;

// Start placeholder time
datetime_t t = {
        .year  = 2000,
        .month = 01,
        .day   = 01,
        .dotw  = 0, // 0 is Sunday, so 5 is Friday
        .hour  = 0,
        .min   = 0,
        .sec   = 00
};

// =======================================
// ntp setup
// =======================================
// The folloqing IP network operations
// are all called by the ntp thread below
// =======================================

typedef struct NTP_T_ {
    ip_addr_t ntp_server_address;
    bool dns_request_sent;
    struct udp_pcb *ntp_pcb;
    absolute_time_t ntp_test_time;
    alarm_id_t ntp_resend_alarm;
} NTP_T;

#define NTP_SERVER "pool.ntp.org"
#define NTP_MSG_LEN 48
#define NTP_PORT 123
#define NTP_DELTA 2208988800 // seconds between 1 Jan 1900 and 1 Jan 1970
#define NTP_TEST_TIME (10 * 1000) //30*1000
#define NTP_RESEND_TIME (10 * 1000)

//========================================
// Called with results of operation
static void ntp_result(NTP_T* state, int status, time_t *result) {
    if (status == 0 && result) {
        struct tm *utc = gmtime(result);
        printf("ntp response: %02d/%02d/%04d %02d:%02d:%02d\n", utc->tm_mday, utc->tm_mon + 1, utc->tm_year + 1900,
               utc->tm_hour, utc->tm_min, utc->tm_sec);
        //printf("wday %d\n", utc->tm_wday); // day of the week
        if(set_rtc_flag){
            set_rtc_flag = 0 ;
            t.year = utc->tm_year + 1900 ;
            t.month = utc->tm_mon + 1 ;
            t.day = utc->tm_mday ;
            t.dotw = utc->tm_wday ;
            t.hour = utc->tm_hour;
            t.min = utc->tm_min;
            t.sec = utc->tm_sec;
            rtc_set_datetime(&t);
        }
    }
    else if (set_rtc_flag){
        set_rtc_flag = 0 ;
        printf("RTC set from NTP failed");

    }

    if (state->ntp_resend_alarm > 0) {
        cancel_alarm(state->ntp_resend_alarm);
        state->ntp_resend_alarm = 0;
    }
    state->ntp_test_time = make_timeout_time_ms(NTP_TEST_TIME);
    state->dns_request_sent = false;
}

// =======================================
static int64_t ntp_failed_handler(alarm_id_t id, void *user_data);

// Make an NTP request
static void ntp_request(NTP_T *state) {
    // cyw43_arch_lwip_begin/end should be used around calls into lwIP to ensure correct locking.
    // You can omit them if you are in a callback from lwIP. Note that when using pico_cyw_arch_poll
    // these calls are a no-op and can be omitted, but it is a good practice to use them in
    // case you switch the cyw43_arch type later.
    cyw43_arch_lwip_begin();
    struct pbuf *p = pbuf_alloc(PBUF_TRANSPORT, NTP_MSG_LEN, PBUF_RAM);
    uint8_t *req = (uint8_t *) p->payload;
    memset(req, 0, NTP_MSG_LEN);
    req[0] = 0x1b;
    udp_sendto(state->ntp_pcb, p, &state->ntp_server_address, NTP_PORT);
    pbuf_free(p);
    cyw43_arch_lwip_end();
}

// =======================================
static int64_t ntp_failed_handler(alarm_id_t id, void *user_data)
{
    NTP_T* state = (NTP_T*)user_data;
    printf("\nntp request failed\n");
    ntp_result(state, -1, NULL);
    return 0;
}

// =======================================
// Call back with a DNS result
static void ntp_dns_found(const char *hostname, const ip_addr_t *ipaddr, void *arg) {
    NTP_T *state = (NTP_T*)arg;
    if (ipaddr) {
        state->ntp_server_address = *ipaddr;
        printf("ntp address %s\n", ipaddr_ntoa(ipaddr));
        ntp_request(state);
    } else {
        printf("ntp dns request failed\n");
        ntp_result(state, -1, NULL);
    }
}

// =======================================
// NTP data received handler
static void ntp_recv(void *arg, struct udp_pcb *pcb, struct pbuf *p, const ip_addr_t *addr, u16_t port) {
    NTP_T *state = (NTP_T*)arg;
    uint8_t mode = pbuf_get_at(p, 0) & 0x7;
    uint8_t stratum = pbuf_get_at(p, 1);

    // Check the result
    if (ip_addr_cmp(addr, &state->ntp_server_address) && port == NTP_PORT && p->tot_len == NTP_MSG_LEN &&
        mode == 0x4 && stratum != 0) {
        uint8_t seconds_buf[4] = {0};
        pbuf_copy_partial(p, seconds_buf, sizeof(seconds_buf), 40);
        uint32_t seconds_since_1900 = seconds_buf[0] << 24 | seconds_buf[1] << 16 | seconds_buf[2] << 8 | seconds_buf[3];
        uint32_t seconds_since_1970 = seconds_since_1900 - NTP_DELTA;
        time_t epoch = seconds_since_1970;
        ntp_result(state, 0, &epoch);
    } else {
        printf("invalid ntp response\n");
        ntp_result(state, -1, NULL);
    }
    pbuf_free(p);
}

// =======================================
// Perform initialisation
static NTP_T* ntp_init(void) {
    NTP_T *state = calloc(1, sizeof(NTP_T));
    if (!state) {
        printf("failed to allocate state\n");
        return NULL;
    }
    state->ntp_pcb = udp_new_ip_type(IPADDR_TYPE_ANY);
    if (!state->ntp_pcb) {
        printf("failed to create pcb\n");
        free(state);
        return NULL;
    }
    udp_recv(state->ntp_pcb, ntp_recv, state);
    return state;
}

// ==================================================
// Runs ntp test forever
// ==================================================
// ntp thread protothread
// ==================================================
static PT_THREAD (protothread_ntp(struct pt *pt)){
//void run_ntp_test(void) {
    PT_BEGIN(pt);
    static  NTP_T *state ;
    state = ntp_init();
    //if (!state)
     //   return;
    while(true) {
        // wait until signaled by serial thread
        PT_SEM_WAIT(pt, &ntp_go_s) ;
        //
        //state = ntp_init();
        if (absolute_time_diff_us(get_absolute_time(), state->ntp_test_time) < 0 && !state->dns_request_sent) {
            // Set alarm in case udp requests are lost
            state->ntp_resend_alarm = add_alarm_in_ms(NTP_RESEND_TIME, ntp_failed_handler, state, true);

            // cyw43_arch_lwip_begin/end should be used around calls into lwIP to ensure correct locking.
            // You can omit them if you are in a callback from lwIP. Note that when using pico_cyw_arch_poll
            // these calls are a no-op and can be omitted, but it is a good practice to use them in
            // case you switch the cyw43_arch type later.
            cyw43_arch_lwip_begin();
            int err = dns_gethostbyname(NTP_SERVER, &state->ntp_server_address, ntp_dns_found, state);
            cyw43_arch_lwip_end();

            state->dns_request_sent = true;
            if (err == ERR_OK) {
                ntp_request(state); // Cached result
            } else if (err != ERR_INPROGRESS) { // ERR_INPROGRESS means expect a callback
                printf("dns request failed\n");
                ntp_result(state, -1, NULL);
            }

            //free(state);
        }
#if PICO_CYW43_ARCH_POLL
        // if you are using pico_cyw43_arch_poll, then you must poll periodically from your
        // main loop (not from a timer interrupt) to check for Wi-Fi driver or lwIP work that needs to be done.
        cyw43_arch_poll();
        // you can poll as often as you like, however if you have nothing else to do you can
        // choose to sleep until either a specified time, or cyw43_arch_poll() has work to do:
        cyw43_arch_wait_for_work_until(state->dns_request_sent ? at_the_end_of_time : state->ntp_test_time);
#else
        // if you are not using pico_cyw43_arch_poll, then WiFI driver and lwIP work
        // is done via interrupt in the background. This sleep is just an example of some (blocking)
        // work you might be doing.
       
#endif
       PT_YIELD_usec(100000);
     
       //sleep_ms(1000) ;
    }   
    free(state);
    PT_END(pt);
}

// ==================================================
// toggle gpio 2  while alternating with ntp
// this is really just a test of multitasking
// compatability with LWIP
// ==================================================
static PT_THREAD (protothread_toggle2(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    //
     // set up LED gpio 2
     gpio_init(2) ;	
     gpio_set_dir(2, GPIO_OUT) ;
     gpio_put(2, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        //
        // toggle gpio 2
        LED_state = !LED_state ;
        gpio_put(2, LED_state);       
        PT_YIELD_INTERVAL(100) ;
        //
        // NEVER exit while
      } // END WHILE(1)
    PT_END(pt);
} // blink thread

// ==================================================
// toggle cyw43 LED  while alternating with ntp
// this is really just a test of multitasking
// compatability with LWIP
// ==================================================
static PT_THREAD (protothread_toggle_cyw43(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    //
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        //
        // toggle gpio 2
        LED_state = !LED_state ;
        
        // does this need???
        //cyw43_arch_lwip_begin();
        cyw43_arch_gpio_put(CYW43_WL_GPIO_LED_PIN, LED_state);
        //cyw43_arch_lwip_end();
        PT_YIELD_INTERVAL(100000) ;
        //
        // NEVER exit while
      } // END WHILE(1)
    PT_END(pt);
} // blink thread

static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
        static char cmd[16], arg1[16], arg2[16], arg3[16], arg4[16], arg5[16], arg6[16] ;
        static char* token ;
      //
      printf("Type 'help' for commands\n\r") ;

      while(1) {
        PT_YIELD_usec(100000);

        // print prompt
        sprintf(pt_serial_out_buffer, "cmd> ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        //sscanf(pt_serial_in_buffer, "%s %f %f %f %f", arg0, &arg1, arg2, arg3, arg4) ;
        // tokenize
        token = strtok(pt_serial_in_buffer, "  ");
        strcpy(cmd, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg1, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg2, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg3, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg4, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg5, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg6, token) ;


        // parse by command
        if(strcmp(cmd,"help")==0){
            // filter design commands
            printf("***\n\rget ntp \n\r"); 
            printf("set rtc\n\r");  
            printf("get rtc\n\r");  
            printf("get local\n\r"); 
            printf("set zone timezone\n") ;
        }
        //
        else if(strcmp(cmd,"get")==0){
            if(strcmp(arg1,"ntp")==0)
                PT_SEM_SIGNAL(pt, &ntp_go_s) ; 
            else  if(strcmp(arg1,"rtc")==0){
                rtc_get_datetime(&t);
                if (t.year == 2000){
                    printf("RTC not set \n");
                }  
                else {
                    datetime_to_str(datetime_str, sizeof(datetime_buf), &t);
                    printf("\r%s UTC\n", datetime_str);
                }  
            }
            //
            else  if(strcmp(arg1,"local")==0){
                // convert to local time
                rtc_get_datetime(&t);
                t.hour += zone ;
                if(t.hour < 0) {
                    t.hour += 24 ;
                    t.day -= 1;
                    t.dotw -= 1 ;
                }
                datetime_to_str(datetime_str, sizeof(datetime_buf), &t);
                printf("\r%s zone=%2d\n", datetime_str, zone);
            }
            //
            else  printf("Huh?\n\r") ;
        }

        //
        else if(strcmp(cmd,"set")==0){
             if(strcmp(arg1,"rtc")==0){
                set_rtc_flag = 1;
                // get a new time and set the rtc
                PT_SEM_SIGNAL(pt, &ntp_go_s) ;  
                PT_YIELD(pt);
             }
             //
             else if(strcmp(arg1,"zone")==0){
                sscanf(arg2, "%d", &zone);
             }
             //
             else  printf("Huh?\n\r") ;
        }
          
          //
        

        // no valid command
        else printf("Huh?\n\r") ;

        // NEVER exit while
      } // END WHILE(1)

  PT_END(pt);
} // serial thread

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock -- if you want
  //set_sys_clock_khz(250000, true); // 171us

  //========================================
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.1.2 two-core\n\r");
  
  //========================================
  // realtime clock init
  rtc_init();

  //========================================
  // network init
  if (cyw43_arch_init()) {
        printf("failed to initialise--exiting\n");
        return 1;
    }

    cyw43_arch_enable_sta_mode();

    if (cyw43_arch_wifi_connect_timeout_ms(WIFI_SSID, WIFI_PASSWORD, CYW43_AUTH_WPA2_AES_PSK, 10000)) {
        printf("failed to connect--will try again\n");
        if (cyw43_arch_wifi_connect_timeout_ms(WIFI_SSID, WIFI_PASSWORD, CYW43_AUTH_WPA2_AES_PSK, 30000)) {
            printf("failed to connect--exiting\n");
            return 1;
        }
    }

    // init pt strucdtures
    PT_SEM_INIT(&ntp_go_s, 1) ;

    //========================================
  // start core 1 threads -- none here
  //multicore_reset_core1();
  //multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  printf("Starting threads\n") ;
  pt_add_thread(protothread_ntp);
  pt_add_thread(protothread_toggle2);
  pt_add_thread(protothread_toggle_cyw43) ;
  pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////